
import human_id
import pandas as pd
from gitlab2pandas.core import Core

class Processing(Core):
   
    def replace_user_id(self):
        """
        Replace user_ids with a pseudonym geneerated by human-id.
        There might be some user_ids in commits that are not connected to a User.

        """
        users_df = self.get_pandas_data_frame(Core.Features.USERS)
        if users_df is None or users_df.empty:
            return
        uuids = []
        users = {}
        for index, row in users_df.iterrows():
            uuid = human_id.generate_id(seed=row["web_url"])
            uuids.append(uuid)
            users[row["id"]] = uuid
        users_df = users_df.assign(uuid=uuids)
        self.save_as_pandas(self.Features.USERS,users_df)
        for filename in self.Features.to_list():
            if filename != "Users":
                self.__try_replace_id_with_uuid(filename, users, "user_id")
                self.__try_replace_id_with_uuid(filename, users, "author_id")
                self.__try_replace_id_with_uuid(filename, users, "owner_id")
                self.__try_replace_id_with_uuid(filename, users, "assignee_id")
                self.__try_replace_id_with_uuid(filename, users, "closed_by_id")
                self.__try_replace_id_with_uuid(filename, users, "merge_user_id")
                self.__try_replace_id_with_uuid(filename, users, "resolved_by_id")
                self.__try_replace_list_id_with_uuid(filename, users, "assignees_ids")
                self.__try_replace_list_id_with_uuid(filename, users, "reviewers_ids")
    
    def __try_replace_id_with_uuid(self, filename, users, column):
        """
        Replaces all user_ids for a standard column in a pandas file with a human-id.
        
        """
        df = self.get_pandas_data_frame(filename)
        if df is None:
            return
        if column in df:
            for user_id, user_uuid in users.items():
                df.loc[df[column] == user_id, column] = user_uuid
        self.save_as_pandas(filename,df)
        
    def __try_replace_list_id_with_uuid(self, filename, users, column):
        """
        Replaces all user_ids for a list column in a pandas file with a human-id.
        
        """
        df = self.get_pandas_data_frame(filename)
        if df is None:
            return
        data_list = []
        if column in df:
            for index, row in df.iterrows():
                uuids = []
                for id in row[column]:
                    if id in users:
                        uuids.append(users[id])
                data = row
                data[column] = uuids
                data_list.append(data)
            self.save_as_pandas(filename,pd.DataFrame(data_list))

    


from gitlab2pandas.core import Core
from gitlab2pandas.extractions import Extractions

class GitLab2Pandas(Core):

    def extract_data(self, extract_parallel:bool = False, feature_blacklist:list = [], feature_whitelist:list = [], update:bool = True) -> None:
        """    
        Extracts GitLab data based on the feature black- or whitelist
        Parallel extraction might fail for some GitLab Server because of server settings.

        Parameters
        ----------
        extract_parallel : bool, default=False
            Extracting the data parallel.
        feature_blacklist : list, default=[]
            Features which will be ignored.
        feature_whitelist : list, default=[]
            Features which will be extracted. If its empty then all features are extracted which are not in the blacklist.
        update: bool, default=True
            Extract only new items after last extration.
            
        """
        extractions = Extractions(self.data_root_dir, 
                                    project=self.project,
                                    extract_parallel=extract_parallel)
        if self.input_file_type != self.FileTypes.PANDAS:
            extractions.set_input_type(self.input_file_type)
        if self.output_file_type != self.FileTypes.PANDAS:
            extractions.set_output_type(self.output_file_type)
        extractions.start(feature_blacklist, feature_whitelist, update)


import unittest
from . import utility
from gitlab2pandas.gitlab2pandas import GitLab2Pandas

class TestUtitlity(unittest.TestCase):

    def test_extract_all_to_json(self):
        gitlab2pandas = GitLab2Pandas(utility.DATA_ROOT_DIR, 
                                        project_namespace = utility.project_namespace(), 
                                        project_name = utility.project_name())
        utility.connect(gitlab2pandas)
        gitlab2pandas.set_input_type(gitlab2pandas.FileTypes.JSON)
        gitlab2pandas.set_output_type(gitlab2pandas.FileTypes.JSON)
        gitlab2pandas.extract_data(True,update=True)
        gitlab2pandas.convert_to_excel("Extraction_data_json")

if __name__ == "__main__":
    unittest.main()


import unittest
from . import utility
from gitlab2pandas.extractions import Extractions

class TestBasics(unittest.TestCase):

    def test_extract_branches(self):
        extractions = Extractions(utility.DATA_ROOT_DIR, 
                                    project_namespace = utility.project_namespace(), 
                                    project_name = utility.project_name())
        utility.connect(extractions)
        extractions.extract_branches()
        pass

    def test_extract_releases(self):
        extractions = Extractions(utility.DATA_ROOT_DIR,
                                    project_namespace = utility.project_namespace(), 
                                    project_name = utility.project_name())
        utility.connect(extractions)
        extractions.extract_releases()
        pass

    def test_extract_pipelines(self):
        extractions = Extractions(utility.DATA_ROOT_DIR,
                                    project_namespace = utility.project_namespace(), 
                                    project_name = utility.project_name())
        utility.connect(extractions)
        extractions.extract_pipelines()
        pass
    
    def test_extract_jobs(self):
        extractions = Extractions(utility.DATA_ROOT_DIR, 
                                    project_namespace = utility.project_namespace(), 
                                    project_name = utility.project_name())
        utility.connect(extractions)
        extractions.extract_jobs()
        pass

    def test_issues(self):
        extractions = Extractions(utility.DATA_ROOT_DIR, 
                                    project_namespace = utility.project_namespace(), 
                                    project_name = utility.project_name())
        utility.connect(extractions)
        extractions.extract_issues()
        pass

    def test_merge_requests(self):
        extractions = Extractions(utility.DATA_ROOT_DIR, 
                                    project_namespace = utility.project_namespace(), 
                                    project_name = utility.project_name())
        utility.connect(extractions)
        extractions.extract_merge_requests()
        pass

    def test_commits(self):
        extractions = Extractions(utility.DATA_ROOT_DIR, 
                                    project_namespace = utility.project_namespace(), 
                                    project_name = utility.project_name())
        utility.connect(extractions)
        extractions.extract_commits()
        pass

    def test_project(self):
        extractions = Extractions(utility.DATA_ROOT_DIR, 
                                    project_namespace = utility.project_namespace(), 
                                    project_name = utility.project_name())
        utility.connect(extractions)
        extractions.extract_project()
        pass
    
if __name__ == "__main__":
    unittest.main()


import unittest
from . import utility
from gitlab2pandas.processing import Processing

class TestBasics(unittest.TestCase):

    def test_replace_user_id(self):
        processing = Processing(utility.DATA_ROOT_DIR, project_namespace=utility.project_namespace(), project_name=utility.project_name())
        processing.replace_user_id()
        processing.convert_to_excel("Processed")
        pass
    
if __name__ == "__main__":
    unittest.main()


from pathlib import Path
import pickle
import pandas as pd
from typing import Union

from gitlab.v4.objects import Project
from gitlab import Gitlab

class Core():
    """
    Initializes core object with general information.
    Decide wheather to initialize with a project object or with the project namespace and name.
    Extractions can only be done with a project object or after connecting to a server with the project namespace and name.

    Parameters
    ----------
    data_root_dir : str
        A existing top level directory for data extraction.
    project : Project, default=None
        Project object from gitlab.
    project_namespace : str, default=None
        Namespace of the project.
    project_name : str, default=None
        Name of the project.

    """
    
    class Features():
        # Name <= 31 chars
        USERS = "Users"
        BRANCHES = "Branches"
        RELEASES = "Releases"
        PIPELINES = "Pipelines"
        PIPELINES_REPORT = "PipelinesReport"
        PIPELINES_BRIDGES = "PipelinesBridges"
        JOBS = "Jobs"
        ISSUES = "Issues"
        ISSUES_NOTES = "IssuesNotes"
        ISSUES_AWARD_EMOJIS = "IssuesAwardEmojis"
        ISSUES_NOTES_AWARD_EMOJIS = "IssuesNotesAwardEmojis"
        ISSUES_RESOURCESTATEEVENTS = "IssuesResourcestateevents"
        ISSUES_RESOURCELABELEVENTS = "IssuesResourcelabelevents"
        ISSUES_RESOURCEMILESTONESEVENTS = "IssuesResourcemilestonesevents"
        ISSUES_CLOSED_BY_MR = "IssuesClosedByMR"
        ISSUES_RELATED_MR = "IssuesRelatedMR"
        ISSUES_LINKS = "IssuesLinks"
        MERGE_REQUESTS = "MergeRequests"
        MERGE_REQUESTS_NOTES = "MRsNotes"
        MERGE_REQUESTS_COMMITS = "MRsCommits"
        MERGE_REQUESTS_AWARD_EMOJIS = "MRsAwardEmojis"
        MERGE_REQUESTS_NOTES_AWARD_EMOJIS = "MRsNotesAwardEmojis"
        MERGE_REQUESTS_RESOURCESTATEEVENTS = "MRsResourcestateevents"
        MERGE_REQUESTS_RESOURCELABELEVENTS = "MRsResourcelabelevents"
        MERGE_REQUESTS_CHANGES = "MRsChanges"
        MERGE_REQUESTS_DIFFS = "MRsDiffs"
        MERGE_REQUESTS_RESOURCEMILESTONESEVENTS = "MRsResourcemilestonesevents"
        COMMITS = "Commits"
        COMMITS_COMMENTS = "CommitsComments"
        COMMITS_REFS = "CommitsRefs"
        COMMITS_DIFFS = "CommitsDiffs"
        COMMITS_STATUSES = "CommitStatuses"
        PROJECTS = "Projects"
        EVENTS = "Events"
        ISSUE_BOARDS = "IssueBoards"
        ISSUE_BOARDS_LISTS = "IssueBoardsLists"
        LABELS = "Labels"
        TRIGGERS = "Triggers"
        PIPELINE_SCHEDULES = "PipelineSchedules"
        RUNNERS = "Runners"
        RUNNERS_JOBS = "RunnersJobs"
        SNIPPETS = "Snippets"
        WIKIS = "Wikis"
        MILESTONES = "Milestones"

        @classmethod
        def to_list(cls) -> list:
            """
            Returns a list of strings with all Features.
            
            Returns
            -------
            list
                A list of strings with all Features.

            """
            features = []
            for var, value in vars(cls).items():
                if isinstance(value,str):
                    if not var.startswith("__"):
                        features.append(value)
            return features

    class FileTypes():
        PANDAS = ".p"
        JSON = ".json"

    def __init__(self, data_root_dir:str, project:Project = None, project_namespace:str = None, project_name:str = None) -> None:
        """
        Initializes core object with general information.
        Decide wheather to initialize with a project object or with the project namespace and name.
        Extractions can only be done with a project object.
        ToDo: log_level=logging.INFO

        Parameters
        ----------
        data_root_dir : str
            A existing top level directory for data extraction.
        project : Project, default=None
            Project object from gitlab.
        project_namespace : str, default=None
           Namespace of the project.
        project_name : str, default=None
            Name of the project.
    
        """
        self.data_root_dir = data_root_dir
        self.project = project
        self.project_namespace = project_namespace
        self.project_name = project_name
        if project is None and (project_namespace is None or project_name is None):
            raise Exception("Need a project or its namespace and name")
        if project is None:
            self.project_data_dir = Path(self.data_root_dir,project_namespace,project_name)
            self.project_data_dir.mkdir(parents=True, exist_ok=True)
        else:
            self.project_namespace = project.attributes["namespace"]["path"]
            self.project_name = project.attributes["path"]
            self.project_data_dir = Path(self.data_root_dir,project.attributes["path_with_namespace"])
            self.project_data_dir.mkdir(parents=True, exist_ok=True)
        self.input_file_type = self.FileTypes.PANDAS
        self.output_file_type = self.FileTypes.PANDAS

    def connect(self, server_url:str, private_token:str=None, oauth_token:str=None, job_token:str=None) -> None:
        """
        Get the project object from GitLab and using the project namespace and name. Only public projects can be accessed (read-only) without a token.
        Extraction can be done after a connection.

        Parameters
        ----------
        server_url: str
            Url to the GitLab server.
        private_token : str, default=None
            Private token or personal token for authentication.
        project_name : str, default=None
            Oauth token for authentication
        project_name : str, default=None
            Job token for authentication (to be used in CI).
            
        """
        gitlab_object = None
        if private_token:
            gitlab_object = Gitlab(server_url, private_token=private_token, per_page=100)
        elif oauth_token:
            gitlab_object = Gitlab(server_url, oauth_token=oauth_token, per_page=100)
        elif job_token:
            gitlab_object = Gitlab(server_url, job_token=job_token, per_page=100)
        else:
            # anonymous gitlab instance, read-only for public resources
            gitlab_object = Gitlab(server_url)
        self.project = gitlab_object.projects.get(f"{self.project_namespace}/{self.project_name}", per_page=100)

    def save_as_pandas(self, filename:str, data:pd.DataFrame) -> None:
        """    
        Saves a pandas DataFrame to the project directory.
        The project metadata will be saved in the top level directory with a filename as pandas file.

        Parameters
        ----------
        filename : str
            Name for the file.
        data : pd.DataFrame
            DataFrame to be saved.
    
        """
        if filename == Core.Features.PROJECTS:
            pd_file = Path(self.data_root_dir, filename + self.output_file_type)
        else:
            pd_file = Path(self.project_data_dir, filename + self.output_file_type)
        if self.output_file_type == self.FileTypes.PANDAS:
            with open(pd_file, "wb") as f:
                pickle.dump(data, f)
        elif self.output_file_type == self.FileTypes.JSON:
            data.to_json(pd_file, indent=2)
    
    def get_pandas_data_frame(self, filename:str) -> Union[pd.DataFrame,None]:
        """    
        Get a pandas DataFrame from the project directory.
        The project metadata will be excessed from the top level directory.

        Parameters
        ----------
        filename : str
            Name of the file to import.
    
    
        Returns
        -------
        DataFrame
            Return a DataFrame of the existing file.
        None
            Return None because the file does not exists.

        """
        if filename == Core.Features.PROJECTS:
            pd_file = Path(self.data_root_dir, filename + self.input_file_type)
        else:
            pd_file = Path(self.project_data_dir, filename + self.input_file_type)
        if pd_file.is_file():
            if self.input_file_type == self.FileTypes.PANDAS:
                return pd.read_pickle(pd_file)
            elif self.input_file_type == self.FileTypes.JSON:
                return pd.read_json(pd_file)
        else:
            return None
              
    def get_pandas_data_frame_path(self, filename:str) -> Union[Path,None]:
        """    
        Get a pandas DataFrame path from the project directory.
        The project metadata will be excessed from the top level directory.

        Parameters
        ----------
        filename : str
            Name of the feature to get the file path.
    
    
        Returns
        -------
        Path
            Return a str path of the feature.
        None
            Return None because the file does not exists.

        """
        if filename == Core.Features.PROJECTS:
            pd_file = Path(self.data_root_dir, filename + self.input_file_type)
        else:
            pd_file = Path(self.project_data_dir, filename + self.input_file_type)
        if pd_file.is_file():
            return pd_file
        else:
            return None

    def set_input_type(self, input_file_type:str) -> bool:
        """    
        Set the input type file and check if the file type is supported by gitlab2pandas.
        Input file type is needed for the update feature.

        Parameters
        ----------
        input_file_type : str
            File ending of the desired input type.
    
        Returns
        -------
        bool
            Return if the input file type was changed.
        """
        if input_file_type == ".p" or input_file_type == ".json":
            self.input_file_type = input_file_type
            return True
        else:
            return False
    
    def set_output_type(self, output_file_type:str) -> bool:
        """    
        Set the output type file and check if the file type is supported by gitlab2pandas.
        Output file type is needed for the automatically dataframe storage of the extrations.

        Parameters
        ----------
        output_file_type : str
            File ending of the desired input type.
    
        Returns
        -------
        bool
            Return if the input file type was changed.
        """
        if output_file_type == ".p" or output_file_type == ".json":
            self.output_file_type = output_file_type
            return True
        else:
            return False

    def convert_to_excel(self, excel_filename, features:list = None) -> None:
        """    
        Converts features to an excel file. If no features are passed, then all features will be converted.

        Parameters
        ----------
        excel_filename : str
            Name for the file.
        features : list, default=None
            Features to convert. If no features are passed, then all features will be converted.
            
        """
        writer = pd.ExcelWriter(Path(self.project_data_dir, f'{excel_filename}.xlsx'), engine='xlsxwriter')
        if features is None:
            features = Core.Features.to_list()
        for feature in features:
            df = self.get_pandas_data_frame(feature)
            if df is not None:
                col_times = [col for col in df.columns if any([isinstance(x, pd.Timestamp) for x in df[col]])]
                for col in col_times:
                    df[col] = pd.to_datetime(
                        df[col], infer_datetime_format=True) 
                    df[col] = df[col].dt.tz_localize(None) 
                df.to_excel(writer, sheet_name=feature)
        writer.close()


# use pipenv-setup to generate the requirement list
# https://pypi.org/project/pipenv-setup/

from setuptools import setup
from gitlab2pandas import __version__


with open("README.md", "r") as f:
   long_description = f.read()



setup(
   name="gitlab2pandas",
   version=__version__,
   packages=["gitlab2pandas"],
   license="BSD 2",
   description="gitlab2pandas supports the aggregation of project activities in a GitLab repository and makes them available in pandas dataframes.",
   long_description = long_description,
   long_description_content_type="text/markdown",
   author="Maximilian Karl",
   url="https://gitlab.com/gitlab-learning-analytics/gitlab2pandas",
   download_url="https://gitlab.com/gitlab-learning-analytics/gitlab2pandas/-/archive/main/gitlab2pandas-main.zip",
   keywords=["git", "gitlab", "collaborative code development", "git mining"],
   install_requires=[
      "human-id>=0.2.0",
      "pandas>=1.4.4",
      "python-gitlab>=3.9.0",
      "XlsxWriter>=3.0.3"
   ], 
   classifiers=[
      "Programming Language :: Python :: 3",
      "Operating System :: OS Independent",
   ],
   python_requires=">=3.9"
)



from typing import Union
import json
import sys
import threading
import queue
import pandas as pd
from gitlab2pandas.core import Core
from gitlab.exceptions import GitlabAuthenticationError

class Extractions(Core):
    """
    Initializes extractions object with general information.
    Decide wheather to initialize with a project object or with the project namespace and name.
    Extractions can only be done with a project object or after connecting to a server with the project namespace and name.

    Parameters
    ----------
    data_root_dir : str
        A existing top level directory for data extraction.
    project : Project, default=None
        Project object from gitlab.
    project_namespace : str, default=None
        Namespace of the project.
    project_name : str, default=None
        Name of the project.
    extract_parallel: bool, default=False
        Parallel extraction might fail for some GitLab Server because of server settings.

    """

    EXTRACTIONS_WITHOUT_UPDATE = [
        Core.Features.BRANCHES,
        Core.Features.ISSUE_BOARDS,
        Core.Features.LABELS,
        Core.Features.MILESTONES,
        Core.Features.PROJECTS,
        Core.Features.RELEASES,
        Core.Features.SNIPPETS,
        Core.Features.USERS,
        Core.Features.WIKIS,
        Core.Features.TRIGGERS
    ]
    
    def __init__(self, data_root_dir: str, project=None, project_namespace=None, project_name=None, extract_parallel=False) -> None:
        """
        Initializes a Extractions object with general information.
        Decide wheather to initialize with a project object or with the project namespace and name.
        Extractions can only be started with a project object.
        ToDo: log_level=logging.INFO

        Parameters
        ----------
        data_root_dir : str
            A existing top level directory for data extraction.
        project : Project, default=None
            Project object from gitlab.
        project_namespace : str, default=None
           Namespace of the project.
        project_name : str, default=None
            Name of the project.
        extract_parallel: bool, default=False
            Parallel extraction might fail for some GitLab Server because of server settings.

        """
        super().__init__(data_root_dir, project, project_namespace, project_name)
        self.extract_parallel = extract_parallel
        self.data_queue = queue.Queue()
        self.consumer_thread = threading.Thread(target=self.__gitlab_data_consumer)
        self.log_queue = queue.Queue()
        self.log_serial_thread = threading.Thread(target=self.__log_serial_consumer)
        self.log_parallel_thread = threading.Thread(target=self.__log_parallel_consumer)
        self.use_feature_whitelist = None
        self.feature_list = []
        self.update_date = None

    def start(self, feature_blacklist:list = [], feature_whitelist:list = [], update:bool = True) -> None:
        """
        Starts a extraction with a blacklist or whitelist for features.
        The extraction can start from the last commit date or the entire project.

        Parameters
        ----------
        feature_blacklist : list, default=[]
            Features which will be ignored.
        feature_whitelist : list, default=[]
            Features which will be extracted. If its empty then all features are extracted which are not in the blacklist.
        update: bool, default=True
            Extract only new items after last extration.

        """
        if self.project is None:
            raise Exception("Need a connection (project object)")
        if self.consumer_thread.is_alive():
            raise Exception("Can not extract Data. There is already one extraction running")
        if feature_whitelist != []:
            self.use_feature_whitelist = True
            self.feature_list = feature_whitelist
            if feature_blacklist != []:
                print("Whitelist is used and Blacklist is ignored!")
        elif feature_blacklist != []:
            self.use_feature_whitelist = False
            self.feature_list = feature_blacklist
        else:
            self.use_feature_whitelist = None
        if update:
            # ToDo: Check projects atributes
            projects_df = self.get_pandas_data_frame(self.Features.PROJECTS)
            if projects_df is not None and not projects_df.empty:
                project_df = projects_df.loc[projects_df["path_with_namespace"] == self.project.attributes["path_with_namespace"]]
                if len(project_df) == 1:
                    self.update_date = project_df.iloc[0]["last_activity_at"]
                else:
                    print("there is redundant information in projects pandas file")
        self.consumer_thread.start()
        method_list = [method for method in dir(Extractions) if method.startswith('extract') is True]
        if self.extract_parallel:
            ### parallel ###
            self.log_parallel_thread.start()
            threads = []
            for method in method_list:
                threads.append(threading.Thread(target=getattr(self,method), args=()))
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
        else:
            ### sequential ### 
            self.log_serial_thread.start()
            for method in method_list:
                getattr(self,method)()
        self.data_queue.put((None,None))
        self.consumer_thread.join()
        self.use_feature_whitelist = None
        self.feature_list = []

    def pass_white_black_list(self, feature) -> bool:
        """
        Checks if a feature passes the white- and blacklist.

        Parameters
        ----------
        feature : str
            Feature to be checked.

        Returns
        -------
        bool
            True if the feature can be extracted.
            False if the feature should be ignored.

        """
        if self.use_feature_whitelist is None:
            return True
        elif self.use_feature_whitelist and feature in self.feature_list:
            return True
        elif not self.use_feature_whitelist and feature not in self.feature_list:
            return True
        return False

    def extract_branches(self) -> None:
        """
        Extracts branches from GitLab.
        Check for update does not work.

        """
        if not self.pass_white_black_list(Core.Features.BRANCHES):
            return
        self.__gitlab_data_producer(self.project, Core.Features.BRANCHES, ["branches", "list"])
        
    def extract_commits(self) -> None:
        """
        Extracts commits and its sub features from GitLab.
        Check for update works.
        
        """
        functions = ["commits", "list"]
        sub_functions = {}
        if self.pass_white_black_list(Core.Features.COMMITS_COMMENTS):
            sub_functions[Core.Features.COMMITS_COMMENTS] = ["comments", "list"]
        if self.pass_white_black_list(Core.Features.COMMITS_REFS):
            sub_functions[Core.Features.COMMITS_REFS] = ["refs"]
        if self.pass_white_black_list(Core.Features.COMMITS_DIFFS):
            sub_functions[Core.Features.COMMITS_DIFFS] = ["diff"]
        if self.pass_white_black_list(Core.Features.COMMITS_STATUSES):
            sub_functions[Core.Features.COMMITS_STATUSES] = ["statuses", "list"]
        if not self.pass_white_black_list(Core.Features.COMMITS):
            if not self.use_feature_whitelist:
                # If the feature is on the blacklist then ignore the feature and its subfeatures
                return
            if sub_functions == {}:
                # If the feature is not on the whitelist then ignore the feature only if no subfeature is on the whitelist
                return
        if sub_functions != {}:
            functions = {
                "attr": functions,
                "sub_functions": sub_functions
            }
        self.__gitlab_data_producer(self.project, Core.Features.COMMITS, functions)

    def extract_events(self) -> None:
        """
        Extracts events from GitLab.
        Check for update works.
        
        """
        if not self.pass_white_black_list(Core.Features.EVENTS):
            return
        self.__gitlab_data_producer(self.project, Core.Features.EVENTS, ["events", "list"])

    def extract_issues(self) -> None:
        """
        Extracts issues and its sub features from GitLab.
        Check for update works.
        
        """
        functions = ["issues", "list"]
        sub_functions = {}
        # ignored ["time_stats"] --> already in issue
        # ignored ["participants"] --> already in issue
        if self.pass_white_black_list(Core.Features.ISSUES_NOTES_AWARD_EMOJIS):
            if not self.use_feature_whitelist is False or Core.Features.ISSUES_NOTES not in self.feature_list:
                # If there is not a blacklist or notes are not on the blacklist then add notes award emojis.
                sub_functions[Core.Features.ISSUES_NOTES] = {
                    "attr": ["notes", "list"], 
                    "sub_functions": {Core.Features.ISSUES_NOTES_AWARD_EMOJIS: ["awardemojis", "list"]}
                }
        elif self.pass_white_black_list(Core.Features.ISSUES_NOTES):
            sub_functions[Core.Features.ISSUES_NOTES] = ["notes", "list"]
        if self.pass_white_black_list(Core.Features.ISSUES_AWARD_EMOJIS):
            sub_functions[Core.Features.ISSUES_AWARD_EMOJIS] = ["awardemojis", "list"]
        if self.pass_white_black_list(Core.Features.ISSUES_RESOURCESTATEEVENTS):
            sub_functions[Core.Features.ISSUES_RESOURCESTATEEVENTS] = ["resourcestateevents", "list"]
        if self.pass_white_black_list(Core.Features.ISSUES_RESOURCELABELEVENTS):
            sub_functions[Core.Features.ISSUES_RESOURCELABELEVENTS] = ["resourcelabelevents", "list"]
        if self.pass_white_black_list(Core.Features.ISSUES_CLOSED_BY_MR):
            sub_functions[Core.Features.ISSUES_CLOSED_BY_MR] = ["closed_by"]
        if self.pass_white_black_list(Core.Features.ISSUES_RELATED_MR):
            sub_functions[Core.Features.ISSUES_RELATED_MR] = ["related_merge_requests"]
        if self.pass_white_black_list(Core.Features.ISSUES_LINKS):
            sub_functions[Core.Features.ISSUES_LINKS] = ["links", "list"]
        if self.pass_white_black_list(Core.Features.ISSUES_RESOURCEMILESTONESEVENTS):
            sub_functions[Core.Features.ISSUES_RESOURCEMILESTONESEVENTS] = ["resourcemilestoneevents", "list"]
        if not self.pass_white_black_list(Core.Features.ISSUES):
            if not self.use_feature_whitelist:
                # If the feature is on the blacklist then ignore the feature and its subfeatures
                return
            if sub_functions == {}:
                # If the feature is not on the whitelist then ignore the feature only if no subfeature is on the whitelist
                return
        if sub_functions != {}:
            functions = {
                "attr": functions,
                "sub_functions": sub_functions
            }
        self.__gitlab_data_producer(self.project, Core.Features.ISSUES, functions)

    def extract_issue_boards(self) -> None:
        """
        Extracts issue boards from GitLab.
        Check for update does not work.

        """
        functions = ["boards", "list"]
        sub_functions = {}
        if self.pass_white_black_list(Core.Features.ISSUE_BOARDS_LISTS):
            sub_functions[Core.Features.ISSUE_BOARDS_LISTS] = ["lists", "list"]
        if not self.pass_white_black_list(Core.Features.ISSUE_BOARDS):
            if not self.use_feature_whitelist:
                # If the feature is on the blacklist then ignore the feature and its subfeatures
                return
            if sub_functions == {}:
                # If the feature is not on the whitelist then ignore the feature only if no subfeature is on the whitelist
                return
        if sub_functions != {}:
            functions = {
                "attr": functions,
                "sub_functions": sub_functions
            }
        self.__gitlab_data_producer(self.project, Core.Features.ISSUE_BOARDS, functions)

    def extract_labels(self) -> None:
        """
        Extracts labels from GitLab.
        Check for update does not work.

        """
        if not self.pass_white_black_list(Core.Features.LABELS):
            return
        self.__gitlab_data_producer(self.project, Core.Features.LABELS, ["labels", "list"])

    def extract_merge_requests(self) -> None:
        """
        Extracts merge requests and its sub features from GitLab.
        Check for update works.
        
        """
        functions = ["mergerequests", "list"]
        sub_functions = {}
        # ignore ["pipelines", "list"] --> pipelines can be matched via commit sha
        # ignored ["time_stats"] --> already in mr
        if self.pass_white_black_list(Core.Features.MERGE_REQUESTS_NOTES_AWARD_EMOJIS):
            if not self.use_feature_whitelist is False or Core.Features.MERGE_REQUESTS_NOTES not in self.feature_list:
                # If there is not a blacklist or notes are not on the blacklist then add notes award emojis.
                sub_functions[Core.Features.MERGE_REQUESTS_NOTES] = {
                    "attr": ["notes", "list"], 
                    "sub_functions": {Core.Features.MERGE_REQUESTS_NOTES_AWARD_EMOJIS: ["awardemojis", "list"]}
                }
        elif self.pass_white_black_list(Core.Features.MERGE_REQUESTS_NOTES):
            sub_functions[Core.Features.MERGE_REQUESTS_NOTES] = ["notes", "list"]
        if self.pass_white_black_list(Core.Features.MERGE_REQUESTS_AWARD_EMOJIS):
            sub_functions[Core.Features.MERGE_REQUESTS_AWARD_EMOJIS] = ["awardemojis", "list"]
        if self.pass_white_black_list(Core.Features.MERGE_REQUESTS_COMMITS):
            sub_functions[Core.Features.MERGE_REQUESTS_COMMITS] = ["commits"]
        if self.pass_white_black_list(Core.Features.MERGE_REQUESTS_CHANGES):
            sub_functions[Core.Features.MERGE_REQUESTS_CHANGES] = ["changes"]
        if self.pass_white_black_list(Core.Features.MERGE_REQUESTS_DIFFS):
            sub_functions[Core.Features.MERGE_REQUESTS_DIFFS] = ["diffs", "list"]
        if self.pass_white_black_list(Core.Features.MERGE_REQUESTS_RESOURCESTATEEVENTS):
            sub_functions[Core.Features.MERGE_REQUESTS_RESOURCESTATEEVENTS] = ["resourcestateevents", "list"]
        if self.pass_white_black_list(Core.Features.MERGE_REQUESTS_RESOURCELABELEVENTS):
            sub_functions[Core.Features.MERGE_REQUESTS_RESOURCELABELEVENTS] = ["resourcelabelevents", "list"]
        if self.pass_white_black_list(Core.Features.MERGE_REQUESTS_RESOURCEMILESTONESEVENTS):
            sub_functions[Core.Features.MERGE_REQUESTS_RESOURCEMILESTONESEVENTS] = ["resourcemilestoneevents", "list"]
        if not self.pass_white_black_list(Core.Features.MERGE_REQUESTS):
            if not self.use_feature_whitelist:
                # If the feature is on the blacklist then ignore the feature and its subfeatures
                return
            if sub_functions == {}:
                # If the feature is not on the whitelist then ignore the feature only if no subfeature is on the whitelist
                return
        if sub_functions != {}:
            functions = {
                "attr": functions,
                "sub_functions": sub_functions
            }
        self.__gitlab_data_producer(self.project, Core.Features.MERGE_REQUESTS, functions)
    
    def extract_milestones(self) -> None:
        """
        Extracts milestones from GitLab.
        Check for update does not work.

        """
        if not self.pass_white_black_list(Core.Features.MILESTONES):
            return
        # milestone.issues() --> in issues
        # milestone.merge_requests() --> in merge requests
        self.__gitlab_data_producer(self.project, Core.Features.MILESTONES, ["milestones", "list"]  )

    def extract_pipelines(self) -> None:
        """
        Extracts pipelines and its sub features from GitLab.
        Check for update works.
        If updated, then it will extract jobs, too.

        """
        functions = ["pipelines", "list"]
        sub_functions = {}
        if self.pass_white_black_list(Core.Features.PIPELINES_REPORT):
            sub_functions[Core.Features.PIPELINES_REPORT] = ["test_report", "get"]
        if self.pass_white_black_list(Core.Features.PIPELINES_BRIDGES):
            sub_functions[Core.Features.PIPELINES_BRIDGES] = ["bridges", "list"]
        if self.update_date is not None:
            sub_functions[Core.Features.JOBS] = ["jobs", "list"]
        if not self.pass_white_black_list(Core.Features.PIPELINES):
            if not self.use_feature_whitelist:
                # If the feature is on the blacklist then ignore the feature and its subfeatures
                return
            if sub_functions == {}:
                # If the feature is not on the whitelist then ignore the feature only if no subfeature is on the whitelist
                return
        if sub_functions != {}:
            functions = {
                "attr": functions,
                "sub_functions": sub_functions
            }
        self.__gitlab_data_producer(self.project, Core.Features.PIPELINES, functions)
    
    def extract_triggers(self) -> None:
        """
        Extracts triggers for pipelines from GitLab.
        Check for update does not work.

        """
        if not self.pass_white_black_list(Core.Features.TRIGGERS):
            return
        self.__gitlab_data_producer(self.project, Core.Features.TRIGGERS, ["triggers", "list"])

    def extract_pipeline_schedules(self) -> None:
        """
        Extracts pipeline schedules for pipelines from GitLab.
        Check for update does not work.

        """
        if not self.pass_white_black_list(Core.Features.PIPELINE_SCHEDULES):
            return
        self.__gitlab_data_producer(self.project, Core.Features.PIPELINE_SCHEDULES, ["pipelineschedules", "list"])

    def extract_jobs(self) -> None:
        """
        Extracts jobs from GitLab.
        Check for update works.
        If updated, then jobs will be extract in pipelines.
        
        """
        if self.update_date is not None:
            return
        if not self.pass_white_black_list(Core.Features.JOBS):
            return
        self.__gitlab_data_producer(self.project, Core.Features.JOBS, ["jobs", "list"])

    def extract_project(self) -> None:
        """
        Extracts general project information from GitLab.
        Check for update does not work.

        """
        def try_len(function, **kwargs):
            try:
                obj = function(**kwargs)
            except GitlabAuthenticationError:
                print(f"Token can not access {function}")
                return None
            else:
                return len(obj)

        if not self.pass_white_black_list(Core.Features.PROJECTS):
            return
        try:
            commits = self.project.commits.list(all=True)
            last_commit_date = commits[0].attributes["created_at"]
            commit_count = len(commits)
        except GitlabAuthenticationError:
            print(f"Token can not access self.project.commits.list")
            last_commit_date = None
            commit_count = None
        project_data = self.__get_gitlab_attributes(self.project.attributes)
        project_data.update({
            "contributor_count": try_len(self.project.repository_contributors,all=True),
            "member_count": try_len(self.project.members_all.list,all=True),
            "branch_count": try_len(self.project.branches.list,all=True),
            "commit_count": commit_count,
            "last_commit_date": last_commit_date,
            "labels_count": try_len(self.project.labels.list,all=True),
            "milestone_count": try_len(self.project.milestones.list,all=True),
            "merge_requests_count": try_len(self.project.mergerequests.list,all=True),
            "release_count":  try_len(self.project.releases.list,all=True),
            "issues_count": try_len(self.project.issues.list,all=True)
        })
        projects_df = self.get_pandas_data_frame(Core.Features.PROJECTS)
        if projects_df is None or projects_df.empty:
            self.save_as_pandas(Core.Features.PROJECTS,pd.DataFrame([project_data]))
            return
        projects_df = projects_df[projects_df.id != project_data['id']]
        projects_df = pd.concat([projects_df, pd.DataFrame([project_data])], ignore_index=True, sort=False)
        self.save_as_pandas(Core.Features.PROJECTS,projects_df)

    def extract_releases(self) -> None:
        """
        Extracts releases from GitLab.
        Check for update does not work.

        """
        if not self.pass_white_black_list(Core.Features.RELEASES):
            return
        self.__gitlab_data_producer(self.project, Core.Features.RELEASES, ["releases", "list"])
        
    def extract_snippets(self) -> None:
        """
        Extracts snippets from GitLab.
        Check for update does not work.

        """
        if not self.pass_white_black_list(Core.Features.SNIPPETS):
            return
        self.__gitlab_data_producer(self.project, Core.Features.SNIPPETS, ["snippets", "list"])

    def extract_users(self) -> None:
        """
        Extracts users from GitLab.
        Check for update does not work.

        """
        if not self.pass_white_black_list(Core.Features.USERS):
            return
        self.__gitlab_data_producer(self.project, Core.Features.USERS, ["users", "list"])

    def extract_wikis(self) -> None:
        """
        Extracts wiki pages from GitLab.
        Check for update does not work.

        """
        if not self.pass_white_black_list(Core.Features.WIKIS):
            return
        self.__gitlab_data_producer(self.project, Core.Features.WIKIS, ["wikis", "list"])
 
    def __log_serial_consumer(self) -> None:
        """
        Method for a consumer thread to log the serial process of extracting data.

        """
        totals = {}
        counts = {}
        size = 60
        max_text_length = 30
        while self.consumer_thread.is_alive() or not self.log_queue.empty():
            if not self.log_queue.empty():
                feature, total = self.log_queue.get()
                if feature not in totals:
                    if totals != {}:
                        sys.stdout.write("\n")
                    totals[feature] = total
                    counts[feature] = 0
                else:
                    counts[feature] += 1
                x = int(size*counts[feature]/total)
                sys.stdout.flush()
                text = f" extracting {feature}:"
                while len(text) < max_text_length:
                    text += " "
                sys.stdout.write("%s[%s%s] %i/%i\r" % (text, "#"*x, "."*(size-x), counts[feature], total))
        sys.stdout.write("\n")
        sys.stdout.flush()
    
    def __log_parallel_consumer(self) -> None:
        """
        Method for a consumer thread to log the parallel process of extracting data.

        """
        features = []
        sum = 0
        count = 0
        size = 60
        while self.consumer_thread.is_alive() or not self.log_queue.empty():
            if not self.log_queue.empty():
                feature, total = self.log_queue.get()
                if feature not in features:
                    features.append(feature)
                    sum += total
                else:
                    count += 1
                x = int(size*count/sum)
                sys.stdout.flush()
                sys.stdout.write("%s[%s%s] %i/%i\r" % (f" extracting parallel:         ", "#"*x, "."*(size-x), count, sum))
        sys.stdout.write("\n")
        sys.stdout.flush()

    def __gitlab_data_consumer(self) -> None:
        """
        Method for a consumer thread to collect the extracted data.
        After everything is extracted the data will be saved to pandas files.

        """
        buffer = {}
        while True:
            if not self.data_queue.empty():
                feature_name, gitlab_data = self.data_queue.get()
                if feature_name is None:
                    break
                if feature_name in buffer:
                    buffer[feature_name].append(gitlab_data)
                else:
                    buffer[feature_name] = [gitlab_data]
        # updated on and after the last day of modification
        if self.update_date is None:
            for key, value in buffer.items():
                self.save_as_pandas(key,pd.DataFrame(value))  
        else:
            # ToDo: Check if there are redudant information
            for key, value in buffer.items():
                if key == self.Features.WIKIS or key == self.Features.BRANCHES or key == self.Features.ISSUE_BOARDS or key == self.Features.ISSUE_BOARDS_LISTS or key == self.Features.LABELS or key == self.Features.USERS:
                    self.save_as_pandas(key,pd.DataFrame(value))
                else:
                    feature_df = self.get_pandas_data_frame(key)
                    new_df = pd.concat([pd.DataFrame(value),feature_df], ignore_index=True)
                    if "iid" in new_df:
                        new_df = new_df.drop_duplicates(subset=['iid'])
                    elif "id" in new_df:
                        new_df = new_df.drop_duplicates(subset=['id'])
                    else:
                        print(f"Ids not found in {key}. Might not drop duplicates")
                        new_df = new_df.drop_duplicates()
                    new_df.reset_index()
                    self.save_as_pandas(key,new_df)

    def __gitlab_data_producer(self, gitlab_obj, feature_name:str, value, is_sub_function:bool = False) -> None:
        """
        Method for a producer thread to extract data from GitLab.
        It can run mutiple producer loops in order to extract subfeatures
        After extracting the data is placed in a queue.

        Parameters
        ----------
        gitlab_obj : Any
            A GitLab object which can acquire information. 
            Top level object is the GitLab project object.
        feature_name : str
            The name of the feature which will be extracted.
        value : Any
            The value as dict defines if the feature has subfeatures to extract.
            The value as list defines a list of methods for the feature in order to extract data.
        is_sub_function : bool
            States if this feature is a subfeature and called by main feature.

        """
        if feature_name == Core.Features.ISSUES_CLOSED_BY_MR:
            x = 0
        obj = gitlab_obj
        has_sub_functions = False
        if isinstance(value, dict):
            attrs = value["attr"]
            has_sub_functions = True
        else:
            attrs = value
        for attr in attrs:
            obj = getattr(obj, attr)
        try:
            if self.update_date is None or self.get_pandas_data_frame_path(feature_name) is None:
                gitlab_data_list = obj(all=True)
            else:
                gitlab_data_list = obj(all=True, since=self.update_date, updated_after=self.update_date, after=self.update_date)
        except GitlabAuthenticationError:
            print(f"Token can not access {feature_name}!")
            return

        threads = []
        if isinstance(gitlab_data_list, dict) or hasattr(gitlab_data_list, "attributes"):
            self.__producer_loop(gitlab_data_list, gitlab_obj, feature_name, value, has_sub_functions)
        else:
            if not is_sub_function and len(gitlab_data_list) > 0:
                self.log_queue.put((feature_name,len(gitlab_data_list)))
            for gitlab_data in gitlab_data_list:
                if has_sub_functions and self.extract_parallel:
                        loop_thread = threading.Thread(target=self.__producer_loop, args=(gitlab_data, gitlab_obj, feature_name, value, has_sub_functions,))
                        loop_thread.start()
                        threads.append(loop_thread)
                else:
                    self.__producer_loop(gitlab_data, gitlab_obj, feature_name, value, has_sub_functions)
                if not is_sub_function:
                    self.log_queue.put((feature_name,len(gitlab_data_list)))
        for thread in threads:
            thread.join()
    
    def __producer_loop(self, gitlab_data, gitlab_obj, feature_name:str, value, has_sub_functions:bool) -> None:
        """
        Method for a thread to start a new data producer or to extact the data from the attributes.

        Parameters
        ----------
        gitlab_data : Any
            Includes the extracted data of the gitlab object.
        gitlab_obj : Any
            A GitLab object which can acquire information. 
            Top level object is the GitLab project object.
        feature_name : str
            The name of the feature which will be extracted.
        value : Any
            The value as dict defines if the feature has subfeatures to extract.
            The value as list defines a list of methods for the feature in order to extract data.
        has_sub_function : bool
            States if this feature has a subfeature and need to be called by current feature.
            
        """
        if feature_name == Core.Features.ISSUES_CLOSED_BY_MR or feature_name == Core.Features.ISSUES_RELATED_MR:
            data = {}
            data["issue_iid"] = gitlab_obj.attributes["iid"]
            data["mr_iid"] = gitlab_data["iid"]
            data["project_id"] = gitlab_data["project_id"]
            self.data_queue.put((feature_name,data))
            return
        if feature_name == Core.Features.MERGE_REQUESTS_COMMITS:
            data = {}
            data["mr_iid"] = gitlab_obj.attributes["iid"]
            data["commit_id"] = gitlab_data.attributes["id"]
            data["project_id"] = gitlab_data.attributes["project_id"]
            self.data_queue.put((feature_name,data))
            return
        if isinstance(gitlab_data, dict):
            if "iid" in gitlab_obj.attributes:
                self.__get_gitlab_attributes(gitlab_data, feature_name, gitlab_obj.attributes["iid"])
            elif "id" in gitlab_obj.attributes:
                self.__get_gitlab_attributes(gitlab_data, feature_name, gitlab_obj.attributes["id"])
        else:
            parent_id = None
            if feature_name == Core.Features.MERGE_REQUESTS_COMMITS:
                parent_id = gitlab_obj.attributes["iid"]
            ## ad mr changes
            self.__get_gitlab_attributes(gitlab_data.attributes, feature_name, parent_id)
            if has_sub_functions:
                for feature_name2, value2 in value["sub_functions"].items():
                    self.__gitlab_data_producer(gitlab_data,feature_name2,value2,True)

    def __get_gitlab_attributes(self, gitlab_data, feature_name:str = None, parent_id = None) -> Union[dict,None]:
        """
        Extract the data from a GitLab feature object and apply special information.

        Parameters
        ----------
        gitlab_data : Any
            Includes the extracted data of the gitlab object
        feature_name : str, default=None
            The name of the feature which will be extracted. 
            Only if the feature is defined then it will place the data in the data_queue or otherwise it will return the data.
        parent_id : Any, default=None
            If the festure has a parent then the parent_id will be added to the data.
            
        Returns
        -------
        dict
            The extracted data will be returned because no feature is defined.
        None
            The extracted data will be passed in the data_queue because a feature is defined.

        """
        data = {}
        if parent_id is not None:
            if "MRs" in feature_name:
                data["mr_iid"] = parent_id
            elif "Commits" in feature_name:
                data["commit_id"] = parent_id
            else:
                data["parent_id"] = parent_id
                print(f"{feature_name} has a unknown parent id!")
        for key, value in gitlab_data.items():
            if isinstance(value, dict):
                if key == "commit" and "id" in value:
                    data["commit_id"] = value["id"]
                elif key == "author" and "id" in value:
                    data["author_id"] = value["id"]
                elif key == "user" and "id" in value:
                    data["user_id"] = value["id"]
                elif key == "owner" and "id" in value:
                    data["owner_id"] = value["id"]
                elif key == "assignee" and "id" in value:
                    data["assignee_id"] = value["id"]
                elif key == "closed_by" and "id" in value:
                    data["closed_by_id"] = value["id"]
                elif key == "merged_by" and "id" in value:
                    # deprecated --> merge_user
                    pass
                elif key == "merge_user" and "id" in value:
                    data["merge_user_id"] = value["id"]
                elif key == "resolved_by" and "id" in value:
                    data["resolved_by_id"] = value["id"]
                elif key == "milestone" and "id" in value:
                    data["milestone_id"] = value["id"]
                elif key == "label" and "id" in value:
                    data["label_id"] = value["id"]
                elif key == "pipeline" and "id" in value:
                    data["pipeline_id"] = value["id"]
                elif key == "namespace":
                    for key2, value2 in value.items():
                        data[f"{key}_{key2}"] = value2
                else:
                    data[key] = json.dumps(value)
            elif isinstance(value, list):
                if key == "labels":
                    data[key] = value
                elif key == "assignees":
                    data["assignees_ids"] = []
                    for assignee in value:
                        data["assignees_ids"].append(assignee["id"])
                elif key == "reviewers":
                    data["reviewers_ids"] = []
                    for reviewer in value:
                        data["reviewers_ids"].append(reviewer["id"])
                elif key == "parent_ids":
                    data["parent_ids"] = []
                    for parent_id in value:
                        data["parent_ids"].append(parent_id)
                elif key == "tag_list":
                    # deprecated --> topics
                    pass
                elif key == "topics":
                    data["topics"] = []
                    for topic in value:
                        data["topics"].append(topic)
                else:
                    data[key] = json.dumps(value)
            else:
                data[key] = value
        if feature_name is None:
            return data
        self.data_queue.put((feature_name,data))

